"""Data structures for persistent homology calculations."""

from collections import namedtuple

from itertools import chain

from operator import itemgetter

from torch_topological.utils import nesting_level

import torch


class PersistenceInformation(namedtuple(
            'PersistenceInformation',
            [
                'pairing',
                'diagram',
                'dimension',
            ],
            # Ensures that there is always a dimension specified, albeit an
            # 'incorrect' one.
            defaults=[None]
        )
):
    """Persistence information data structure.

    This is a light-weight data structure for carrying information about
    the calculation of persistent homology. It consists of the following
    components:

    - A *persistence pairing*
    - A *persistence diagram*
    - An (optional) *dimension*

    Due to its lightweight nature, no validity checks are performed, but
    all calculation modules should return a sequence of instances of the
    :class:`PersistenceInformation` class.

    Since this data class is shared with modules that are capable of
    calculating persistent homology, the exact form of the persistence
    pairing might change. Please refer to the respective classes for
    more documentation.
    """

    __slots__ = ()

    # Disable iterating over the class since it collates heterogeneous
    # information and should rather be treated as a building block.
    __iter__ = None


def make_tensor(x):
    """Create dense tensor representation from sparse inputs.

    This function turns sparse inputs of :class:`PersistenceInformation`
    objects into 'dense' tensor representations, thus providing a useful
    integration into differentiable layers.

    The dimension of the resulting tensor depends on maximum number of
    topological features, summed over all dimensions in the data. This
    is similar to the format in `giotto-ph`.

    Parameters
    ----------
    x : list of (list of ...) :class:`PersistenceInformation`
        Input, consisting of a (potentially) nested list of
        :class:`PersistenceInformation` objects as obtained
        from a persistent homology calculation module, such
        as :class:`VietorisRipsComplex`.

    Returns
    -------
    torch.tensor
        Dense tensor representation of `x`. The output is best
        understood by considering some examples: given a batch
        obtained from :class:`VietorisRipsComplex`, our tensor
        will have shape `(B, N, 3)`. `B` is the batch size and
        `N` is the sum of maximum lengths of diagrams relative
        to this batch. Each entry will consist of a creator, a
        destroyer, and a dimension. Dummy entries, used to pad
        the batch, can be detected as `torch.nan`.
    """
    level = nesting_level(x)

    # Internal utility function for calculating the length of the output
    # tensor. This is required to ensure that all inputs can be *merged*
    # into a single output tensor.
    def _calculate_length(x, level):

        # Simple base case; should never occur in practice but let's be
        # consistent here.
        if len(x) == 0:
            return 0

        # `chain.from_iterable()` only removes one layer of nesting, but
        # we may have more.
        elif level > 2:
            x = list(chain.from_iterable(x))

        # Collect information that we need to create the full tensor. An
        # entry of the resulting list contains the length of the diagram
        # and the dimension, making it possible to derive padding values
        # for all entries.
        M = list(map(
            lambda a: (len(a.diagram), a.dimension), chain.from_iterable(x)
        ))

        # Get maximum dimension
        dim = max(M, key=itemgetter(1))[1]

        # Get *sum* of maximum number of entries for each dimension.
        # This is calculated over all batches.
        N = sum([
            max([L for L in M if L[1] == d], key=itemgetter(0))[0]
            for d in range(dim + 1)
        ])

        return N

    N = _calculate_length(x, level)

    # List of lists: the first axis is treated as the batch axis, while
    # the second axis is treated as the dimension of diagrams or pairs.
    if level == 2:
        tensors = [
            make_tensor_from_persistence_information(pers_infos)
            for pers_infos in x
        ]

        # Pad all tensors to length N in the first dimension, then turn
        # them into a batch.
        result = torch.stack(
                list(
                    map(
                        lambda t: torch.nn.functional.pad(
                                t,
                                (0, 0, N - len(t), 0),
                                mode='constant',
                                value=torch.nan),
                        tensors
                    )
                )
        )

        return result

    # List of lists of lists: this indicates image-based data, where we
    # also have a set of tensors for each channel. The internal layout,
    # i.e. our input, has the following structure:
    #
    # B x C x D
    #
    # Each variable being the length of the respective list. We want an
    # output of the following shape:
    #
    # B x C x N x 3
    #
    # Here, `N` is the maximum length of an individual persistence
    # information object.
    else:
        tensors = [
            [
                make_tensor_from_persistence_information(pers_infos)
                for pers_infos in batch
            ]
            for batch in x
        ]

        # Pad all tensors to length N in the first dimension, then turn
        # them into a batch. We first stack over channels (inner), then
        # over the batch (outer).
        result = torch.stack([
            torch.stack(
                list(
                    map(
                        lambda t: torch.nn.functional.pad(
                                t,
                                (0, 0, N - len(t), 0),
                                mode='constant',
                                value=torch.nan),
                        batch_tensors
                    )
                )
            )
            for batch_tensors in tensors
        ])

        return result


def make_tensor_from_persistence_information(
    pers_info,
    extract_generators=False
):
    """Convert (sequence) of persistence information entries to tensor.

    This function converts instance(s) of :class:`PersistenceInformation`
    objects into a single tensor. No padding will be performed. A client
    may specify what type of information to extract from the object. For
    instance, by default, the function will extract persistence diagrams
    but this behaviour can be changed by setting `extract_generators` to
    `true`.

    Parameters
    ----------
    pers_info : :class:`PersistenceInformation` or iterable thereof
        Input persistence information object(s). The function is able to
        handle both single objects and sequences. This has no bearing on
        the length of the returned tensor.

    extract_generators : bool
        If set, extracts generators instead of persistence diagram from
        `pers_info`.

    Returns
    -------
    torch.tensor
        Tensor of shape `(n, 3)`, where `n` is the sum of all features,
        over all dimensions in the input `pers_info`. Each triple shall
        be of the form `(creation, destruction, dim)` for a persistence
        diagram. If the client requested generators to be returned, the
        first two entries of the triple refer to *indices* with respect
        to the input data set. Depending on the algorithm employed, the
        meaning of these indices can change. Please refer to the module
        used to calculate persistent homology for more details.
    """
    # Looks a little bit cumbersome, but since `namedtuple` is iterable
    # as well, we need to ensure that we are actually dealing with more
    # than one instance here.
    if len(pers_info) > 1 and not \
            isinstance(pers_info[0], PersistenceInformation):
        pers_info = [pers_info]

    # TODO: This might not always work since the size of generators
    # changes in different dimensions.
    if extract_generators:
        pairs = torch.cat(
            [torch.as_tensor(x.pairing, dtype=torch.float) for x in pers_info],
        ).long()
    else:
        pairs = torch.cat(
            [torch.as_tensor(x.diagram, dtype=torch.float) for x in pers_info],
        ).float()

    dimensions = torch.cat(
        [
            torch.as_tensor([x.dimension] * len(x.diagram), dtype=torch.long)
            for x in pers_info
        ]
    )

    result = torch.column_stack((pairs, dimensions))
    return result


def batch_iter(x, dim=None):
    """Iterate over batches from input data.

    This utility function simplifies working with 'sparse' data sets
    consisting of :class:`PersistenceInformation` instances. It will
    present inputs in the order in which they appear in a batch such
    that instances belonging to the same data set are kept together.

    Parameters
    ----------
    x : recursively-nested list of :class:`PersistenceInformation`
        Input in sparse form, i.e. a nested structure containing
        persistence information about a data set.

    dim : int or `None`
        If set, only iterates over persistence information instances of
        the specified dimension. Else, will iterate over all instances.

    Returns
    -------
    A generator (iterable) that will either yield direct instances of
    :class:`PersistenceInformation` objects or further iterators into
    them. This ensures that it is possible to always iterate over the
    individual batches, without having to know internal details about
    the structure of `x`.
    """
    level = nesting_level(x)

    # Nothing to do for non-nested data structures, i.e. a single batch
    # that has been squeezed (for instance). Wrapping the input enables
    # us to treat it like a regular input again.
    if level == 1:
        x = [x]

    if level <= 2:
        def handler(x): return x

    # Remove the first dimension but also the subsequent one so that all
    # only iterables containing persistence information about a specific
    # data set are being returned.
    #
    # TODO: Generalise recursively? Do we want to support that?
    else:
        def handler(x): return chain.from_iterable(x)

    if dim is not None:
        for x_ in x:
            yield list(filter(lambda x: x.dimension == dim, handler(x_)))
    else:
        for x_ in x:
            yield handler(x_)



"""Demo for summary statistics minimisation of a point cloud.

This example demonstrates how to use various topological summary
statistics in order to change the shape of an input point cloud.
The script can either demonstrate how to adjust the shape of two
point clouds, i.e. using a summary statistic as a loss function,
or how to change the shape of a *single* point cloud. By default
two point clouds will be used.
"""

import argparse

import matplotlib.pyplot as plt

from torch_topological.data import sample_from_disk
from torch_topological.data import sample_from_unit_cube

from torch_topological.nn import SummaryStatisticLoss
from torch_topological.nn import VietorisRipsComplex

from tqdm import tqdm

import torch
import torch.optim as optim


def main(args):
    """Run example."""
    n = args.n_samples
    n_iterations = args.n_iterations
    statistic = args.statistic
    p = args.p
    q = args.q

    vr = VietorisRipsComplex(dim=2)

    if args.single:
        X = sample_from_unit_cube(n=n, d=2)
        Y = X.clone()
    else:
        X = sample_from_disk(n=n, r=0.5, R=0.6)
        Y = sample_from_disk(n=n, r=0.9, R=1.0)
        pi_target = vr(Y)

    # Make source point cloud adjustable by treating it as a parameter.
    # This enables topological loss functions to influence the shape of
    # `X`.
    X = torch.nn.Parameter(torch.as_tensor(X), requires_grad=True)

    loss_fn = SummaryStatisticLoss(
        summary_statistic=statistic,
        p=p,
        q=q
    )

    opt = optim.SGD([X], lr=0.05)

    progress = tqdm(range(n_iterations))

    for i in progress:
        pi_source = vr(X)

        if not args.single:
            loss = loss_fn(pi_source, pi_target)
        else:
            loss = loss_fn(pi_source)

        opt.zero_grad()
        loss.backward()
        opt.step()

        progress.set_postfix(loss=f'{loss.item():.08f}')

    X = X.detach().numpy()

    if args.single:
        plt.scatter(X[:, 0], X[:, 1], label='Result')
        plt.scatter(Y[:, 0], Y[:, 1], label='Initial')
    else:
        plt.scatter(X[:, 0], X[:, 1], label='Source')
        plt.scatter(Y[:, 0], Y[:, 1], label='Target')

    plt.legend()
    plt.show()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-i', '--n-iterations',
        default=250,
        type=int,
        help='Number of iterations'
    )

    parser.add_argument(
        '-n', '--n-samples',
        default=100,
        type=int,
        help='Number of samples in point clouds'
    )

    parser.add_argument(
        '-s', '--statistic',
        choices=[
            'persistent_entropy',
            'polynomial_function',
            'total_persistence',
        ],
        default='polynomial_function',
        help='Name of summary statistic to use for the loss'
    )

    parser.add_argument(
        '-S', '--single',
        action='store_true',
        help='If set, uses only a single point cloud'
    )

    parser.add_argument(
        '-p',
        type=float,
        default=2.0,
        help='Outer exponent for summary statistic loss calculation'
    )

    parser.add_argument(
        '-q',
        type=float,
        default=2.0,
        help='Inner exponent for summary statistic loss calculation. Will '
             'only be used for certain summary statistics.'
    )

    args = parser.parse_args()
    main(args)



"""Example demonstrating the computation of alpha complexes.

This simple example demonstrates how to use alpha complexes to change
the appearance of a point cloud, following the `TopologyLayer
<https://github.com/bruel-gabrielsson/TopologyLayer>`_ package.

This example is still a **work in progress**.
"""

from torch_topological.nn import AlphaComplex
from torch_topological.nn import SummaryStatisticLoss

import numpy as np
import matplotlib.pyplot as plt

import torch

if __name__ == '__main__':
    np.random.seed(42)
    data = np.random.rand(100, 2)

    alpha_complex = AlphaComplex()

    loss_fn = SummaryStatisticLoss(
        summary_statistic='polynomial_function',
        p=1,
        q=2
    )

    X = torch.nn.Parameter(torch.as_tensor(data), requires_grad=True)
    opt = torch.optim.Adam([X], lr=1e-2)

    for i in range(100):
        loss = loss_fn(alpha_complex(X))

        opt.zero_grad()
        loss.backward()
        opt.step()

    X = X.detach().numpy()

    plt.scatter(X[:, 0], X[:, 1])
    plt.show()



"""Demo for topology-regularised autoencoders.

This example demonstrates how to use `pytorch-topological` to create an
additional differentiable loss term that makes autoencoders aware of
topological features. See [Moor20a]_ for more information.
"""

import torch
import torch.optim as optim

import matplotlib.pyplot as plt

from tqdm import tqdm

from torch.utils.data import DataLoader

from torch_topological.datasets import Spheres

from torch_topological.nn import SignatureLoss
from torch_topological.nn import VietorisRipsComplex


class LinearAutoencoder(torch.nn.Module):
    """Simple linear autoencoder class.

    This module performs simple embeddings based on an MSE loss. This is
    similar to ordinary principal component analysis. Notice that the
    class is only meant to provide a simple example that can be run
    easily even without the availability of a GPU. In practice, there
    are many more architectures with improved expressive power
    available.
    """

    def __init__(self, input_dim, latent_dim=2):
        """Create new autoencoder with pre-defined latent dimension."""
        super().__init__()

        self.input_dim = input_dim
        self.latent_dim = latent_dim

        self.encoder = torch.nn.Sequential(
            torch.nn.Linear(self.input_dim, self.latent_dim)
        )

        self.decoder = torch.nn.Sequential(
            torch.nn.Linear(self.latent_dim, self.input_dim)
        )

        self.loss_fn = torch.nn.MSELoss()

    def encode(self, x):
        """Embed data in latent space."""
        return self.encoder(x)

    def decode(self, z):
        """Decode data from latent space."""
        return self.decoder(z)

    def forward(self, x):
        """Embeds and reconstructs data, returning a loss."""
        z = self.encode(x)
        x_hat = self.decode(z)

        # The loss can of course be changed. If this is your first time
        # working with autoencoders, a good exercise would be to 'grok'
        # the meaning of different losses.
        reconstruction_error = self.loss_fn(x, x_hat)
        return reconstruction_error


class TopologicalAutoencoder(torch.nn.Module):
    """Wrapper for a topologically-regularised autoencoder.

    This class uses another autoencoder model and imbues it with an
    additional topology-based loss term.
    """
    def __init__(self, model, lam=1.0):
        super().__init__()

        self.lam = lam
        self.model = model
        self.loss = SignatureLoss(p=2)

        # TODO: Make dimensionality configurable
        self.vr = VietorisRipsComplex(dim=0)

    def forward(self, x):
        z = self.model.encode(x)

        pi_x = self.vr(x)
        pi_z = self.vr(z)

        geom_loss = self.model(x)
        topo_loss = self.loss([x, pi_x], [z, pi_z])

        loss = geom_loss + self.lam * topo_loss
        return loss


if __name__ == '__main__':
    # We first have to create a data set. This follows the original
    # publication by Moor et al. by introducing a simple 'manifold'
    # data set consisting of multiple spheres.
    n_spheres = 11
    data_set = Spheres(n_spheres=n_spheres)

    train_loader = DataLoader(
        data_set,
        batch_size=32,
        shuffle=True,
        drop_last=True
    )

    # Let's set up the two models that we are training. Note that in
    # a real application, you would have a more complicated training
    # setup, potentially with early stopping etc. This training loop
    # is merely to be seen as a proof of concept.
    model = LinearAutoencoder(input_dim=data_set.dimension)
    topo_model = TopologicalAutoencoder(model, lam=10)

    optimizer = optim.Adam(topo_model.parameters(), lr=1e-3)

    n_epochs = 5

    progress = tqdm(range(n_epochs))

    for i in progress:
        topo_model.train()

        for batch, (x, y) in enumerate(train_loader):
            loss = topo_model(x)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        progress.set_postfix(loss=loss.item())

    # Evaluate the autoencoder on a new instance of the data set.
    data_set = Spheres(
        train=False,
        n_samples=2000,
        n_spheres=n_spheres,
    )

    test_loader = DataLoader(
            data_set,
            shuffle=False,
            batch_size=len(data_set)
    )

    X, y = next(iter(test_loader))
    Z = model.encode(X).detach().numpy()

    plt.scatter(
        Z[:, 0], Z[:, 1],
        c=y,
        cmap='Set1',
        marker='o',
        alpha=0.9,
    )
    plt.show()



"""Utilities module."""

from .general import is_iterable
from .general import nesting_level
from .general import wrap_if_not_iterable

from .summary_statistics import total_persistence
from .summary_statistics import persistent_entropy
from .summary_statistics import polynomial_function

__all__ = [
    'is_iterable',
    'nesting_level',
    'persistent_entropy',
    'polynomial_function',
    'total_persistence',
    'wrap_if_not_iterable',
]



"""Summary statistics for persistence diagrams."""

import torch


def persistent_entropy(D, **kwargs):
    """Calculate persistent entropy of a persistence diagram.

    Parameters
    ----------
    D : `torch.tensor`
        Persistence diagram, assumed to be in shape `(n, 2)`, where each
        entry corresponds to a tuple of the form :math:`(x, y)`, with
        :math:`x` denoting the creation of a topological feature and
        :math:`y` denoting its destruction.

    Returns
    -------
    Persistent entropy of `D`.
    """
    persistence = torch.diff(D)
    persistence = persistence[torch.isfinite(persistence)].abs()

    P = persistence.sum()
    probabilities = persistence / P

    # Ensures that a probability of zero will just result in
    # a logarithm of zero as well. This is required whenever
    # one deals with entropy calculations.
    indices = probabilities > 0
    log_prob = torch.zeros_like(probabilities)
    log_prob[indices] = torch.log2(probabilities[indices])

    return torch.sum(-probabilities * log_prob)


def polynomial_function(D, p, q, **kwargs):
    r"""Parametrise polynomial function over persistence diagrams.

    This function follows an approach by Adcock et al. [Adcock16a]_ and
    parametrises a polynomial function over a persistence diagram.

    Parameters
    ----------
    D : `torch.tensor`
        Persistence diagram, assumed to be in shape `(n, 2)`, where each
        entry corresponds to a tuple of the form :math:`(x, y)`, with
        :math:`x` denoting the creation of a topological feature and
        :math:`y` denoting its destruction.

    p : int
        Exponent for persistence differences in the diagram.

    q : int
        Exponent for mean persistence in the diagram.

    Returns
    -------
    Sum of the form :math:`\sigma L^p * \mu^q`, with :math:`L` denoting
    an individual persistence value, and :math:`\mu` denoting its
    average persistence.

    References
    ----------
    .. [Adcock16a] A. Adcock et al., "The Ring of Algebraic Functions on
        Persistence Bar Codes", *Homology, Homotopy and Applications*,
        Volume 18, Issue 1, pp. 381--402, 2016.
    """
    lengths = torch.diff(D)
    means = torch.sum(D, dim=-1, keepdim=True) / 2

    # Filter out non-finite values; the same mask works here because the
    # mean is non-finite if and only if the persistence is.
    mask = torch.isfinite(lengths)
    lengths = lengths[mask]
    means = means[mask]

    return torch.sum(torch.mul(lengths.pow(p), means.pow(q)))


def total_persistence(D, p=2, **kwargs):
    """Calculate total persistence of a persistence diagram.

    This function will calculate the totla persistence of a persistence
    diagram. Infinite value will be ignored.

    Parameters
    ----------
    D : `torch.tensor`
        Persistence diagram, assumed to be in shape `(n, 2)`, where each
        entry corresponds to a tuple of the form :math:`(x, y)`, with
        :math:`x` denoting the creation of a topological feature and
        :math:`y` denoting its destruction.

    p : float
        Weight parameter for the total persistence calculation.

    Returns
    -------
    Total persistence of `D`.
    """
    persistence = torch.diff(D)
    persistence = persistence[torch.isfinite(persistence)]

    return persistence.abs().pow(p).sum()



"""Vietoris--Rips complex calculation module(s)."""

from gph import ripser_parallel
from torch import nn

from torch_topological.nn import PersistenceInformation

import torch


class VietorisRipsComplex(nn.Module):
    """Calculate Vietoris--Rips complex of a data set.

    This module calculates 'differentiable' persistence diagrams for
    point clouds. The underlying topological approximations are done
    by calculating a Vietoris--Rips complex of the data.
    """

    def __init__(self, dim=1, p=2, **kwargs):
        """Initialise new module.

        Parameters
        ----------
        dim : int
            Calculates persistent homology up to (and including) the
            prescribed dimension.

        p : float
            Exponent for the `p`-norm calculation of distances.

        **kwargs
            Additional arguments to be provided to ``ripser``, i.e. the
            backend for calculating persistent homology. The `n_threads`
            parameter, which controls parallelisation, is probably the
            most relevant parameter to be adjusted.
            Please refer to the `the gitto-ph documentation
            <https://giotto-ai.github.io/giotto-ph/build/html/index.html>`_
            for more details on admissible parameters.

        Notes
        -----
        This module currently only supports Minkowski norms. It does not
        yet support other metrics.
        """
        super().__init__()

        self.dim = dim
        self.p = p

        # Ensures that the same parameters are used whenever calling
        # `ripser`.
        self.ripser_params = {
            'return_generators': True,
            'maxdim': self.dim,
        }

        self.ripser_params.update(kwargs)

    def forward(self, x):
        """Implement forward pass for persistence diagram calculation.

        The forward pass entails calculating persistent homology on
        a point cloud and returning a set of persistence diagrams.

        Parameters
        ----------
        x : array_like
            Input point cloud(s). `x` can either be a 2D array of shape
            `(n, d)`, which is treated as a single point cloud, or a 3D
            array/tensor of the form `(b, n, d)`, with `b` representing
            the batch size. Alternatively, you may also specify a list,
            possibly containing point clouds of non-uniform sizes.

        Returns
        -------
        list of :class:`PersistenceInformation`
            List of :class:`PersistenceInformation`, containing both the
            persistence diagrams and the generators, i.e. the
            *pairings*, of a certain dimension of topological features.
            If `x` is a 3D array, returns a list of lists, in which the
            first dimension denotes the batch and the second dimension
            refers to the individual instances of
            :class:`PersistenceInformation` elements.

            Generators will be represented in the persistence pairing
            based on vertex--edge pairs (dimension 0) or edge--edge
            pairs. Thus, the persistence pairing in dimension zero will
            have three components, corresponding to a vertex and an
            edge, respectively, while the persistence pairing for higher
            dimensions will have four components.
        """
        # Check whether individual batches need to be handled (3D array)
        # or not (2D array). We default to this type of processing for a
        # list as well.
        if isinstance(x, list) or len(x.shape) == 3:

            # TODO: This is rather ugly and inefficient but it is the
            # easiest workaround for now.
            return [
                self._forward(torch.as_tensor(x_)) for x_ in x
            ]
        else:
            return self._forward(torch.as_tensor(x))

    def _forward(self, x):
        """Handle a *single* point cloud.

        This internal function handles the calculation of topological
        features for a single point cloud, i.e. an `array_like` of 2D
        shape.

        Parameters
        ----------
        x : array_like of shape `(n, d)`
            Single input point cloud.

        Returns
        -------
        list of class:`PersistenceInformation`
            List of persistence information data structures, containing
            the persistence diagram and the persistence pairing of some
            dimension in the input data set.
        """
        generators = ripser_parallel(
            x.cpu().detach(),
            **self.ripser_params
        )['gens']

        # TODO: Is this always required? Can we calculate this in
        # a smarter fashion?
        distances = torch.cdist(x, x, p=self.p)

        # We always have 0D information.
        persistence_information = \
            self._extract_generators_and_diagrams(
                distances,
                generators,
                dim0=True,
            )

        # Check whether we have any higher-dimensional information that
        # we should return.
        if self.dim >= 1:
            persistence_information.extend(
                self._extract_generators_and_diagrams(
                    distances,
                    generators,
                    dim0=False,
                )
            )

        return persistence_information

    def _extract_generators_and_diagrams(
            self,
            dist,
            gens,
            finite=True,
            dim0=False
    ):
        """Extract generators and persistence diagrams from raw data.

        This convenience function translates between the output of
        `ripser_parallel` and the required output of this function.
        """
        index = 1 if not dim0 else 0
        gens = gens[index]

        # TODO: Handling of infinite features not provided yet, but the
        # index shift is already correct.
        if not finite:
            index += 1

        if dim0:
            # In a Vietoris--Rips complex, all vertices are created at
            # time zero.
            creators = torch.zeros_like(torch.as_tensor(gens)[:, 0])
            destroyers = dist[gens[:, 1], gens[:, 2]]

            persistence_diagram = torch.stack(
                (creators, destroyers), 1
            )

            return [PersistenceInformation(gens, persistence_diagram, 0)]
        else:
            result = []

            for index, gens_ in enumerate(gens):
                creators = dist[gens_[:, 0], gens_[:, 1]]
                destroyers = dist[gens_[:, 2], gens_[:, 3]]

                persistence_diagram = torch.stack(
                    (creators, destroyers), 1
                )

                # Dimension zero is handled differently, so we need to
                # use an offset here.
                dimension = index + 1

                result.append(
                    PersistenceInformation(
                        gens_,
                        persistence_diagram,
                        dimension)
                )

        return result



"""Example of topology-based GANs."""

import torch
import torchvision

from torch_topological.nn import Cubical
from torch_topological.nn import SummaryStatisticLoss

from tqdm import tqdm

import numpy as np
import matplotlib.pyplot as plt


class Generator(torch.nn.Module):
    def __init__(self, latent_dim, shape):
        super().__init__()

        self.latent_dim = latent_dim
        self.output_dim = np.prod(shape)

        self.shape = shape

        def _make_layer(input_dim, output_dim):
            layers = [
                torch.nn.Linear(input_dim, output_dim),
                torch.nn.BatchNorm1d(output_dim, 0.8),
                torch.nn.LeakyReLU(0.2)
            ]
            return layers

        self.model = torch.nn.Sequential(
            *_make_layer(self.latent_dim, 64),
            *_make_layer(64, 128),
            *_make_layer(128, 256),
            torch.nn.Linear(256, self.output_dim),
            torch.nn.Sigmoid()
        )

    def forward(self, z):
        point_cloud = self.model(z)
        point_cloud = point_cloud.view(point_cloud.size(0), *self.shape)

        return point_cloud


class Discriminator(torch.nn.Module):
    def __init__(self, shape):
        super().__init__()

        input_dim = np.prod(shape)

        # Inspired by the original GAN. THERE CAN ONLY BE ONE!
        self.model = torch.nn.Sequential(
            torch.nn.Linear(input_dim, 512),
            torch.nn.LeakyReLU(0.2),
            torch.nn.Linear(512, 256),
            torch.nn.LeakyReLU(0.2),
            torch.nn.Linear(256, 1),
            torch.nn.Sigmoid(),
        )

    def forward(self, x):
        # Flatten point cloud
        x = x.view(x.size(0), -1)
        return self.model(x)


class TopologicalAdversarialLoss(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.cubical = Cubical()
        self.loss = SummaryStatisticLoss('polynomial_function', p=2, q=2)

    def forward(self, real, synthetic):

        loss = 0.0

        for x, y in zip(real, synthetic):
            x = x.squeeze()
            y = y.squeeze()

            pi_real = self.cubical(x)[0]
            pi_synthetic = self.cubical(y)[0]

            loss += self.loss([pi_real], [pi_synthetic])

        return loss


def show(imgs):
    if not isinstance(imgs, list):
        imgs = [imgs]
    fix, axs = plt.subplots(ncols=len(imgs), squeeze=False)
    for i, img in enumerate(imgs):
        img = img.detach()
        img = torchvision.transforms.functional.to_pil_image(img)
        axs[0, i].imshow(np.asarray(img))
        axs[0, i].set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])


if __name__ == '__main__':

    n_epochs = 5

    img_size = 16
    shape = (1, img_size, img_size)
    batch_size = 32
    latent_dim = 200

    data_loader = torch.utils.data.DataLoader(
        torchvision.datasets.MNIST(
            "./data/MNIST",
            train=False,
            download=True,
            transform=torchvision.transforms.Compose(
                [
                    torchvision.transforms.Resize(img_size),
                    torchvision.transforms.ToTensor(),
                    torchvision.transforms.Normalize([0.5], [0.5])
                ]
            ),
        ),
        batch_size=batch_size,
        shuffle=True,
    )

    generator = Generator(shape=shape, latent_dim=latent_dim)
    discriminator = Discriminator(shape=shape)
    adversarial_loss = torch.nn.BCELoss()
    topo_loss = TopologicalAdversarialLoss()

    opt_g = torch.optim.Adam(generator.parameters(), lr=1e-4)
    opt_d = torch.optim.Adam(discriminator.parameters(), lr=1e-4)

    for epoch in range(n_epochs):
        for batch, (imgs, _) in tqdm(enumerate(data_loader), desc='Batch'):
            z = torch.autograd.Variable(
                torch.Tensor(
                    np.random.normal(
                        0,
                        1,
                        (imgs.shape[0], latent_dim)
                    )
                )
            )

            real_labels = torch.as_tensor([1.0] * len(imgs)).view(-1, 1)
            fake_labels = torch.as_tensor([0.0] * len(imgs)).view(-1, 1)

            opt_g.zero_grad()

            imgs_synthetic = generator(z)

            generator_loss = adversarial_loss(
                discriminator(imgs_synthetic), real_labels
            ) + 0.01 * topo_loss(imgs, imgs_synthetic)

            generator_loss.backward()

            opt_g.step()

            opt_d.zero_grad()

            real_loss = adversarial_loss(discriminator(imgs), real_labels)
            fake_loss = adversarial_loss(
                discriminator(imgs_synthetic).detach(), fake_labels
            )

            discriminator_loss = 0.5 * (real_loss + fake_loss)
            discriminator_loss.backward()

            opt_d.step()

    output = imgs_synthetic.detach()
    grid = torchvision.utils.make_grid(output)

    show(grid)

    plt.show()



"""Alpha complex calculation module(s)."""

from torch import nn

from torch_topological.nn import PersistenceInformation

import gudhi
import itertools
import torch


class AlphaComplex(nn.Module):
    """Calculate persistence diagrams of an alpha complex.

    This module calculates persistence diagrams of an alpha complex,
    i.e. a subcomplex of the Delaunay triangulation, which is sparse
    and thus often substantially smaller than other complex.

    It was first described in [Edelsbrunner94]_ and is particularly
    useful when analysing low-dimensional data.

    Notes
    -----
    At the moment, this alpha complex implementation, following other
    implementations, provides *distance-based filtrations* only. This
    means that the resulting persistence diagrams do *not* correspond
    to the circumradius of a simplex.

    In addition, this implementation is **work in progress**. Some of
    the core features, such as handling of infinite features, are not
    available at the moment.

    References
    ----------
    .. [Edelsbrunner94] H. Edelsbrunner and E.P. Mücke,
       "Three-dimensional alpha shapes", *ACM Transactions on Graphics*,
       Volume 13, Number 1, pp. 43--72, 1994.
    """

    def __init__(self, p=2):
        """Initialise new alpha complex calculation module.

        Parameters
        ----------
        p : float
            Exponent for the `p`-norm calculation of distances.

        Notes
        -----
        This module currently only supports Minkowski norms. It does not
        yet support other metrics.
        """
        super().__init__()

        self.p = p

    def forward(self, x):
        """Implement forward pass for persistence diagram calculation.

        The forward pass entails calculating persistent homology on
        a point cloud and returning a set of persistence diagrams.

        Parameters
        ----------
        x : array_like
            Input point cloud(s). `x` can either be a 2D array of shape
            `(n, d)`, which is treated as a single point cloud, or a 3D
            array/tensor of the form `(b, n, d)`, with `b` representing
            the batch size. Alternatively, you may also specify a list,
            possibly containing point clouds of non-uniform sizes.

        Returns
        -------
        list of :class:`PersistenceInformation`
            List of :class:`PersistenceInformation`, containing both the
            persistence diagrams and the generators, i.e. the
            *pairings*, of a certain dimension of topological features.
            If `x` is a 3D array, returns a list of lists, in which the
            first dimension denotes the batch and the second dimension
            refers to the individual instances of
            :class:`PersistenceInformation` elements.

            Generators will be represented in the persistence pairing
            based on proper creator--destroyer pairs of simplices. In
            dimension `k`, for instance, every generator is stored as
            a `k`-simplex followed by a `k+1` simplex.
        """
        # TODO: Copied from `VietorisRipsComplex`; needs to be
        # refactored into a unified interface.
        #
        # Check whether individual batches need to be handled (3D array)
        # or not (2D array). We default to this type of processing for a
        # list as well.
        if isinstance(x, list) or len(x.shape) == 3:

            # TODO: This is rather ugly and inefficient but it is the
            # easiest workaround for now.
            return [
                self._forward(torch.as_tensor(x_)) for x_ in x
            ]
        else:
            return self._forward(torch.as_tensor(x))

    def _forward(self, x):
        alpha_complex = gudhi.alpha_complex.AlphaComplex(
            x.cpu().detach(),
            precision='fast',
        )

        st = alpha_complex.create_simplex_tree()
        st.persistence()
        persistence_pairs = st.persistence_pairs()

        max_dim = x.shape[-1]
        dist = torch.cdist(x.contiguous(), x.contiguous(), p=self.p)

        return [
            self._extract_generators_and_diagrams(
                dist,
                persistence_pairs,
                dim
            )
            for dim in range(0, max_dim)
        ]

    def _extract_generators_and_diagrams(self, dist, persistence_pairs, dim):
        pairs = [
            torch.cat((torch.as_tensor(p[0]), torch.as_tensor(p[1])), 0)
            for p in persistence_pairs if len(p[0]) == (dim + 1)
        ]

        # TODO: Ignore infinite features for now. Will require different
        # handling in the future.
        pairs = [p for p in pairs if len(p) == 2 * dim + 3]

        if not pairs:
            return PersistenceInformation(
                    pairing=[],
                    diagram=[],
                    dim=dim
            )

        # Create tensor of shape `(n, 2 * dim + 3)`, with `n` being the
        # number of finite persistence pairs.
        pairs = torch.stack(pairs)

        # We have to branch here because the creation of
        # zero-dimensional persistence diagrams is easy,
        # whereas higher-dimensional diagrams require an
        # involved lookup strategy.
        if dim == 0:
            creators = torch.zeros_like(torch.as_tensor(pairs)[:, 0])

        # Iterate over the flag complex in order to get (a) the distance
        # of the creator simplex, and (b) the distance of the destroyer.
        # We *cannot* look this information up in the filtration itself,
        # because we have to use gradient-imbued information such as the
        # set of pairwise distances.
        else:
            creators = torch.stack(
                    [
                        self._get_filtration_weight(creator, dist)
                        for creator in pairs[:, :dim+1]
                    ]
            )

        # For the destroyers, we can always rely on the same
        # construction, regardless of dimensionality.
        destroyers = torch.stack(
                [
                    self._get_filtration_weight(destroyer, dist)
                    for destroyer in pairs[:, dim+1:]
                ]
        )

        # Create the persistence diagram from creator and destroyer
        # information. This step is the same for all dimensions.
        persistence_diagram = torch.stack(
            (creators, destroyers), 1
        )

        return PersistenceInformation(
                pairing=pairs,
                diagram=persistence_diagram,
                dimension=dim
        )

    def _get_filtration_weight(self, simplex, dist):
        """Auxiliary function for querying simplex weights.

        This function returns the filtration weight of an arbitrary
        simplex under a distance-based filtration, i.e. the maximum
        weight of its cofaces. This function is crucial for getting
        persistence diagrams that are differentiable.

        Parameters
        ----------
        simplex : torch.tensor
            Simplex to query; must be a sequence of numbers, i.e. of
            shape `(n, 1)` or of shape `(n, )`.

        dist : torch.tensor
            Matrix of pairwise distances between points.

        Returns
        -------
        torch.tensor
            Scalar tensor containing the filtration weight.
        """
        weights = torch.stack([
            dist[edge] for edge in itertools.combinations(simplex, 2)
        ])

        # TODO: Might have to be adjusted depending on filtration
        # ordering?
        return torch.max(weights)



"""Cubical complex calculation module."""

from torch import nn

from torch_topological.nn import PersistenceInformation

import gudhi
import torch

import numpy as np


class CubicalComplex(nn.Module):
    """Calculate cubical complex persistence diagrams.

    This module calculates 'differentiable' persistence diagrams for
    structured data, such as images. This is achieved by calculating
    a *cubical complex*.

    Cubical complexes are the natural choice for calculating topological
    features of highly-structured inputs. See [Rieck20a]_ for an example
    of how to apply such topological features in practice.

    References
    ----------
    .. [Rieck20a] B. Rieck et al., "Uncovering the Topology of
       Time-Varying fMRI Data Using Cubical Complex", *Advances in
       Neural Information Processing Systems 33*, pp. 6900--6912, 2020.
    """

    # TODO: Handle different dimensions?
    def __init__(self, superlevel=False, dim=None):
        """Initialise new module.

        Parameters
        ----------
        superlevel : bool
            Indicates whether to calculate topological features based on
            superlevel sets. By default, *sublevel set filtrations* are
            used.

        dim : int or `None`
            If set, describes dimension of input data. This is meant to
            be the dimension of an individual image **without** channel
            information, if any. The value of `dim` will change the way
            an input tensor is being handled: additional dimensions, if
            present, will be treated as batches or channels. If not set
            to an integer value, :func:`forward` will just *guess* what
            to do with an input (which should work in most cases).
        """
        super().__init__()

        # TODO: This is handled somewhat inelegantly below. Might be
        # smarter to update.
        self.superlevel = superlevel
        self.dim = dim

    def forward(self, x):
        """Implement forward pass for persistence diagram calculation.

        The forward pass entails calculating persistent homology on a
        cubical complex and returning a set of persistence diagrams.
        The way the input will be interpreted depends on the presence
        of the `dim` attribute of this class. If `dim` is set, the
        *last* `dim` dimensions of an input tensor will be considered to
        contain the image data. If `dim` is not set, image dimensions
        will be guessed as follows:

        1. Tensor of `dim = 2`: a single 2D image
        2. Tensor of `dim = 3`: a single 2D image with channels
        3. Tensor of `dim = 4`: a batch of 2D images with channels

        See parameters for more details.

        Parameters
        ----------
        x : array_like
            Input image(s). If `dim` has not been set, will *guess* how
            to handle the input as follows: `x` can either be a 2D array
            of shape `(H, W)`, which is treated as a single image, or
            a 3D array/tensor of the form `(C, H, W)`, with `C`
            representing the number of channels, or a 4D array/tensor of
            the form `(B, C, H, W)`, with `B` being the batch size. If
            `dim` has been set, the same handling strategy applies, but
            the *last* `dim` dimensions of the tensor are being used for
            the cubical complex calculation. All subsequent dimensions
            will be assumed to represent batches or channels (in this
            order). Hence, if `dim` is set, the tensor must at most have
            `dim + 2` dimensions.

        Returns
        -------
        list of :class:`PersistenceInformation`
            List of :class:`PersistenceInformation`, containing both the
            persistence diagrams and the generators, i.e. the
            *pairings*, of a certain dimension of topological features.
            If `x` is a 3D array, returns a list of lists, in which the
            first dimension denotes the batch and the second dimension
            refers to the individual instances of
            :class:`PersistenceInformation` elements. Similar for
            higher-order tensors.

        """
        # Dimension was provided; this makes calculating the *effective*
        # dimension of the tensor much easier: take everything but the
        # last `self.dim` dimensions.
        if self.dim is not None:
            shape = x.shape[:-self.dim]
            dims = len(shape)

        # No dimension was provided; just use the shape provided by the
        # client.
        else:
            dims = len(x.shape) - 2

        # No additional dimensions present: a single image
        if dims == 0:
            return self._forward(x)

        # Handle image with channels, such as a tensor of the form `(C, H, W)`
        elif dims == 1:
            return [
                self._forward(x_) for x_ in x
            ]

        # Handle image with channels and batch index, such as a tensor of
        # the form `(B, C, H, W)`.
        elif dims == 2:
            return [
                    [self._forward(x__) for x__ in x_] for x_ in x
            ]

    def _forward(self, x):
        """Handle a single-channel image.

        This internal function handles the calculation of topological
        features for a single-channel image, i.e. an `array_like`.

        Parameters
        ----------
        x : array_like of shape `(d_1, d_2, ..., d_d)`
            Single-channel input image of arbitrary dimensions. Batch
            dimensions and channel dimensions have to to be handled by
            the calling function explicitly. This function interprets
            its input as a high-dimensional image.

        Returns
        -------
        list of class:`PersistenceInformation`
            List of persistence information data structures, containing
            the persistence diagram and the persistence pairing of some
            dimension in the input data set.
        """
        if self.superlevel:
            x = -x

        cubical_complex = gudhi.CubicalComplex(
            dimensions=x.shape,
            top_dimensional_cells=x.flatten()
        )

        # We need the persistence pairs first, even though we are *not*
        # using them directly here.
        cubical_complex.persistence()
        cofaces = cubical_complex.cofaces_of_persistence_pairs()

        max_dim = len(x.shape)

        # TODO: Make this configurable; is it possible that users only
        # want to return a *part* of the data?
        persistence_information = [
            self._extract_generators_and_diagrams(
                x,
                cofaces,
                dim
            ) for dim in range(0, max_dim)
        ]

        return persistence_information

    def _extract_generators_and_diagrams(self, x, cofaces, dim):
        pairs = torch.empty((0, 2), dtype=torch.long)

        try:
            regular_pairs = torch.as_tensor(
                cofaces[0][dim], dtype=torch.long
            )
            pairs = torch.cat(
                (pairs, regular_pairs)
            )
        except IndexError:
            pass

        try:
            infinite_pairs = torch.as_tensor(
                cofaces[1][dim], dtype=torch.long
            )
        except IndexError:
            infinite_pairs = None

        if infinite_pairs is not None:
            # 'Pair off' all the indices
            max_index = torch.argmax(x)
            fake_destroyers = torch.empty_like(infinite_pairs).fill_(max_index)

            infinite_pairs = torch.stack(
                (infinite_pairs, fake_destroyers), 1
            )

            pairs = torch.cat(
                (pairs, infinite_pairs)
            )

        return self._create_tensors_from_pairs(x, pairs, dim)

    # Internal utility function to handle the 'heavy lifting:'
    # creates tensors from sets of persistence pairs.
    def _create_tensors_from_pairs(self, x, pairs, dim):

        xs = x.shape

        # Notice that `creators` and `destroyers` refer to pixel
        # coordinates in the image.
        creators = torch.as_tensor(
                np.column_stack(
                    np.unravel_index(pairs[:, 0], xs)
                ),
                dtype=torch.long
        )
        destroyers = torch.as_tensor(
                np.column_stack(
                    np.unravel_index(pairs[:, 1], xs)
                ),
                dtype=torch.long
        )
        gens = torch.as_tensor(torch.hstack((creators, destroyers)))

        # TODO: Most efficient way to generate diagram again?
        persistence_diagram = torch.stack((
            x.ravel()[pairs[:, 0]],
            x.ravel()[pairs[:, 1]]
        ), 1)

        return PersistenceInformation(
                pairing=gens,
                diagram=persistence_diagram,
                dimension=dim
        )



"""Data sets based on simple topological shapes."""

from torch_topological.data import sample_from_sphere
from torch_topological.data import sample_from_torus

from torch.utils.data import Dataset

import torch


class SphereVsTorus(Dataset):
    """Data set containing point cloud samples from spheres and tori."""

    def __init__(
        self,
        n_point_clouds=500,
        n_samples=100,
        shuffle=True
    ):
        """Create new instance of the data set.

        Parameters
        ----------
        n_point_clouds : int
            Number of point clouds to generate. Each point cloud will
            consist of `n_samples` samples.

        n_samples : int
            Number of samples to use for each of the `n_point_clouds`
            point clouds contained in the data set.

        shuffle : bool
            If set, shuffles point clouds. Else, point clouds will be
            stored in the order of their creation.
        """
        self.n_point_clouds = n_point_clouds
        self.n_samples = n_samples

        n_spheres = n_point_clouds // 2
        n_tori = n_point_clouds - n_spheres

        spheres = torch.stack([
            sample_from_sphere(self.n_samples) for i in range(n_spheres)
        ])

        tori = torch.stack([
            sample_from_torus(self.n_samples) for i in range(n_tori)
        ])

        labels = torch.as_tensor(
            [0] * n_spheres + [1] * n_tori,
            dtype=torch.long
        )

        self.data = torch.vstack((spheres, tori))
        self.labels = labels

        if shuffle:
            perm = torch.randperm(self.n_point_clouds)

            self.data = self.data[perm]
            self.labels = self.labels[perm]

    def __getitem__(self, index):
        """Get point cloud at `index`.

        Accesses the point cloud stored at `index` and returns it as
        well as its corresponding label.

        Parameters
        ----------
        index : int
            Index of samples to access.

        Returns
        -------
        Tuple of torch.tensor, torch.tensor
            Point cloud at index `index` and its label.
        """
        return self.data[index], self.labels[index]

    def __len__(self):
        """Get number of point clouds stored in data set.

        Returns
        -------
        int
            Number of point clouds stored in this instance of the class.
        """
        return len(self.data)



"""Distance calculation modules between topological descriptors."""

import ot
import torch

from torch_topological.utils import wrap_if_not_iterable


class WassersteinDistance(torch.nn.Module):
    """Implement Wasserstein distance between persistence diagrams.

    This module calculates the Wasserstein between two persistence
    diagrams. The Wasserstein distance is arguably the most common
    metric that is applied when dealing with such diagrams. Notice
    that calculating the metric involves solving optimal transport
    problems, which are known to suffer from scalability problems.
    When dealing with large persistence diagrams, other losses may
    be more appropriate.
    """

    def __init__(self, p=torch.inf, q=1):
        """Create new Wasserstein distance calculation module.

        Parameters
        ----------
        p : float or `inf`
            Specifies the exponent of the norm to calculate. By default,
            `p = torch.inf`, corresponding to the *maximum norm*.

        q: float
            Specifies the order of Wasserstein metric to calculate. This
            raises all internal matching costs to the power of `q`, hence
            subsequently returning the `q`-th root of the total cost.
        """
        super().__init__()

        self.p = p
        self.q = q

    def _project_to_diagonal(self, diagram):
        x = diagram[:, 0]
        y = diagram[:, 1]

        # TODO: Is this the closest point in all p-norms?
        return 0.5 * torch.stack(((x + y), (x + y)), 1)

    def _distance_to_diagonal(self, diagram):
        return torch.linalg.vector_norm(
            diagram - self._project_to_diagonal(diagram),
            self.p,
            dim=1
        )

    def _make_distance_matrix(self, D1, D2):
        dist_D11 = self._distance_to_diagonal(D1)
        dist_D22 = self._distance_to_diagonal(D2)

        # n x m matrix containing the distances between 'regular'
        # persistence pairs of both persistence diagrams.
        dist = torch.cdist(D1, D2, p=torch.inf)

        # Extend the matrix with a column of distances of samples in D1
        # to their respective projection on the diagonal.
        upper_blocks = torch.hstack((dist, dist_D11[:, None]))

        # Create a lower row of distances of samples in D2 to their
        # respective projection on the diagonal. The ordering needs
        # to follow the ordering of samples in D2. Note how one `0`
        # needs to be added to the row in order to balance it. The
        # entry intuitively describes the cost between *projected*
        # points, so it has to be zero.
        lower_blocks = torch.cat(
            (dist_D22, torch.tensor(0, device=dist_D22.device).unsqueeze(0))
        )

        # Full (n + 1 ) x (m + 1) matrix containing *all* distances. By
        # construction, M[[i, n] contains distances to projected points
        # in D1, whereas M[m, j] does the same for points in D2. Only a
        # cell M[i, j] with 0 <= i < n and 0 <= j < m contains a proper
        # distance.
        M = torch.vstack((upper_blocks, lower_blocks))
        M = M.pow(self.q)

        return M

    def forward(self, X, Y):
        """Calculate Wasserstein metric based on input tensors.

        Parameters
        ----------
        X : list or instance of :class:`PersistenceInformation`
            Topological features of the first space. Supposed to contain
            persistence diagrams and persistence pairings.

        Y : list or instance of :class:`PersistenceInformation`
            Topological features of the second space. Supposed to
            contain persistence diagrams and persistence pairings.

        Returns
        -------
        torch.tensor
            A single scalar tensor containing the distance between the
            persistence diagram(s) contained in `X` and `Y`.
        """
        total_cost = 0.0

        X = wrap_if_not_iterable(X)
        Y = wrap_if_not_iterable(Y)

        for pers_info in zip(X, Y):
            D1 = pers_info[0].diagram
            D2 = pers_info[1].diagram

            n = len(D1)
            m = len(D2)

            dist = self._make_distance_matrix(D1, D2)

            # Create weight vectors. Since the last entries of entries
            # describe the m points coming from D2, we have to set the
            # last entry accordingly.

            a = torch.ones(n + 1, device=dist.device)
            b = torch.ones(m + 1, device=dist.device)

            a[-1] = m
            b[-1] = n

            # TODO: Make settings configurable?
            total_cost += ot.emd2(a, b, dist)

        return total_cost.pow(1.0 / self.q)
